#include <stdlib.h>
#include <math.h>
#include "catmull_rom.h"

static int Alpha;
static double tj(double ti, Points_t pi, Points_t pj);
static void catmull_rom_spline(Points_t *inpoints, int num_points, Points_t *points);

/*
   converted from the Python example in 
   https://en.wikipedia.org/wiki/Centripetal_Catmull%E2%80%93Rom_spline#Code_example_in_Python

   Compute the points in the spline segment
   :param inpoints[4]: The (x,y) point pairs that define the Catmull-Rom spline
   :param num_points: The number of points to include in the resulting curve segment
   :points: return: The points
   points must be allocated by caller, number of elements >= num_points
   */

void catmull_rom_spline(Points_t *inpoints, int num_points, Points_t *points)
{
   Points_t P0, P1, P2, P3;

   P0 = inpoints[0];
   P1 = inpoints[1];
   P2 = inpoints[2];
   P3 = inpoints[3];

   double t0 = 0;
   double t1 = tj(t0,P0,P1);
   double t2 = tj(t1,P1,P2);
   double t3 = tj(t2,P2,P3);

   double dt = (t2-t1)/(num_points-1);

   for (int i=0; i<num_points; i++)
   {
      Points_t A1,A2,A3,B1,B2;
      double t = t1 + dt*i;
      A1.x =        (t1 - t) / (t1 - t0) * P0.x + (t - t0) / (t1 - t0) * P1.x;
      A1.y =        (t1 - t) / (t1 - t0) * P0.y + (t - t0) / (t1 - t0) * P1.y;
      A2.x =        (t2 - t) / (t2 - t1) * P1.x + (t - t1) / (t2 - t1) * P2.x;
      A2.y =        (t2 - t) / (t2 - t1) * P1.y + (t - t1) / (t2 - t1) * P2.y;
      A3.x =        (t3 - t) / (t3 - t2) * P2.x + (t - t2) / (t3 - t2) * P3.x;
      A3.y =        (t3 - t) / (t3 - t2) * P2.y + (t - t2) / (t3 - t2) * P3.y;
      B1.x =        (t2 - t) / (t2 - t0) * A1.x + (t - t0) / (t2 - t0) * A2.x;
      B1.y =        (t2 - t) / (t2 - t0) * A1.y + (t - t0) / (t2 - t0) * A2.y;
      B2.x =        (t3 - t) / (t3 - t1) * A2.x + (t - t1) / (t3 - t1) * A3.x;
      B2.y =        (t3 - t) / (t3 - t1) * A2.y + (t - t1) / (t3 - t1) * A3.y;
      points[i].x = (t2 - t) / (t2 - t1) * B1.x + (t - t1) / (t2 - t1) * B2.x;
      points[i].y = (t2 - t) / (t2 - t1) * B1.y + (t - t1) / (t2 - t1) * B2.y;
   }
}

double tj(double ti, Points_t pi, Points_t pj)
{
   if(Alpha == 0)
      return ti + 1.0;
   double dx = pj.x - pi.x;
   double dy = pj.y - pi.y;
   double l = sqrt(dx*dx + dy*dy);
   switch(Alpha)
   {
      case 1:        // alpha = 0.5
	 return ti + sqrt(l);
	 break;
      default:       // alpha = 1.0
	 return ti + l;
   }
   // the original formula: return ti + pow(l,alpha);
}

/*
   Calculate Catmull-Rom for a sequence of initial points and return the combined curve.
   :param base_points: Base points from which the quadruples for the algorithm are taken
   :param num_points: The number of points to include in each curve segment
   :chain: return: The chain of all points (points of all segments)
   :alpha: 1 for the centripetal spline, 0 for the uniform spline, others for the chordal spline.
      resp:  alpha = 0.5, 0.0, 1.0 in the original formula 
   */
void catmull_rom_chain(Points_t *base_points, int n,  int num_points, Points_t *chain, int alpha)
{
   Alpha = alpha;
   Points_t *points = (Points_t*) malloc(num_points*sizeof(Points_t));

   int l = 0;
   for(int i=0; i < n-3; i+=4)
   {
      for(int j=0; j<4; j++)
      {
	 catmull_rom_spline(&base_points[i+j], num_points, points);
	 for(int k=0; k<num_points; k++)
	    chain[l+k] = points[k];
	 l += num_points;
      }
   }
   free(points);
}
